unit B2MConv;

// b2mconv.pas
//
//     Copyright(c) 2009-2011 PT2K <PT2K@d8.dion.ne.jp>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Dialogs, Graphics,
  ResStrs;

const
  MaxBitmapHeight = 75; // o-c ~ o9g
  MaxColorMapSize = 17; // 16 colors + transparent color
  MaxPaletteSize = 16;

type
  TNoteType = (ntx0, ntx1, ntx2);
const
  NoteTypeNames: array[TNoteType] of String = ('x0', 'x1', 'x2');

var
  optTransparentColorSet, optHelpSet, optNoteLengthSet, optOutputFileSet,
  optTopOffsetSet, optTempoSet, optStdoutSet, optNoUseTransparentColorSet,
  optUseTransparentColorSet, optVersionSet, optVerboseSet, optNoWizardSet,
  optWithWizardSet, optNoteTypeSet, optBitmapPathSet, optBitmapPathsSet,
  optErrorSet: Boolean;
  optTransparentColorArg, optNoteLengthArg, optOutputFileArg, optTopOffsetArg,
  optTempoArg, optNoteTypeArg, optBitmapPathArg: String;

var
  Verbose: Boolean;
  UseWizard: Boolean;
  BitmapPath: TFileName;
  SourceBitmap: TBitmap;
  ColorMap: array[0..MaxColorMapSize - 1] of TColor;
  ColorMapSize: Integer;
  UseTransparentColor: Boolean;
  TransparentColorIndex: Integer;
  Palette: array[0..MaxPaletteSize - 1] of TColor;
  PaletteSize: Integer;
  MaxOffset: Integer;
  Offset: Integer;
  NoteType: TNoteType;
  NoteLength: String;
  Tempo: Integer;
  MusData: TStringList;
  MusPath: TFileName;

type
  TB2MProgressEvent = procedure(Progress: Integer) of Object;
  TB2MProgressProc = procedure(Progress: Integer);
var
  OnB2MProgress: TB2MProgressEvent;
  B2MProgressProc: TB2MProgressProc;

procedure DoError(ConsoleMsg, DialogMsg: String);
procedure DoWarning(ConsoleMsg, DialogMsg: String);
procedure DoInformation(ConsoleMsg, DialogMsg: String);
function SetBitmapPath(ABitmapPath: TFileName): Boolean;
function ForceUseTransparentColor(): Boolean;
procedure SetPalette();
function CanSetOffset(): Boolean;
procedure SetOffset(AOffset: Integer);
procedure BitmapToMus();
procedure SaveMus();
procedure WriteMus();

implementation

procedure DoMessage(var F: Text; DlgType: TMsgDlgType; ConsoleMsg, DialogMsg: String);
begin
{$IFDEF WINDOWS}
  MessageDlg(DialogMsg, DlgType, [mbOK], 0);
{$ELSE}
  if UseWizard then
    MessageDlg(DialogMsg, DlgType, [mbOK], 0)
  else
    Writeln(F, ConsoleMsg);
{$ENDIF}
end;

procedure DoError(ConsoleMsg, DialogMsg: String);
begin
  DoMessage(StdErr, mtError, 'Error: ' + ConsoleMsg, 'Error: ' + DialogMsg);
end;

procedure DoWarning(ConsoleMsg, DialogMsg: String);
begin
  if Verbose then
    DoMessage(StdErr, mtWarning, 'Warning: ' + ConsoleMsg, 'Warning: ' + DialogMsg);
end;

procedure DoInformation(ConsoleMsg, DialogMsg: String);
begin
  DoMessage(StdOut, mtInformation, ConsoleMsg, DialogMsg);
end;

function SetBitmapPath(ABitmapPath: TFileName): Boolean;
  function SetColorMap(): Boolean;
  var
    x, y, i: Integer;
    C: TColor;
    Found: Boolean;
  begin
    Result := False;
    ColorMapSize := 0;
    for x := 0 to SourceBitmap.Width - 1 do
      for y := 0 to SourceBitmap.Height - 1 do
      begin
        C := SourceBitmap.Canvas.Pixels[x, y];
        Found := False;
        for i := 0 to ColorMapSize - 1 do
          if ColorMap[i] = C then
          begin
            Found := True;
            break;
          end;
        if not Found then
        begin
          Inc(ColorMapSize);
          if MaxColorMapSize < ColorMapSize then
          begin
            DoError(msgETooManyColors, resETooManyColors);
            exit;
          end;
          ColorMap[ColorMapSize - 1] := C;
        end;
      end;
    if ForceUseTransparentColor() then
      UseTransparentColor := True;
    Result := True;
  end;
begin
  Result := False;
  if BitmapPath = ABitmapPath then
  begin
    Result := True;
    exit;
  end;
  BitmapPath := '';
  MaxOffset := 0;
  if not FileExistsUTF8(ABitmapPath) then
  begin
    DoError(msgEFileNotFound, resEFileNotFound);
    exit;
  end;
  try
    SourceBitmap.LoadFromFile(ABitmapPath);
  except
    on exception do
    begin
      DoError(msgEWrongImageFormat, resEWrongImageFormat);
      exit;
    end;
  end;
  if MaxBitmapHeight < SourceBitmap.Height then
  begin
    DoError(msgEImageTooHigh, resEImageTooHigh);
    exit;
  end;
  MaxOffset := MaxBitmapHeight - SourceBitmap.Height;
  if (Offset < 0) or (MaxOffset < Offset) then
    Offset := 0;
  if not SetColorMap then exit;
  BitmapPath := ABitmapPath;
  MusPath := ChangeFileExt(BitmapPath, '.mus');
  if optOutputFileSet then
    MusPath := optOutputFileArg;
  Result := True;
end;

function ForceUseTransparentColor(): Boolean;
begin
  Result := MaxColorMapSize = ColorMapSize;
end;

procedure SetPalette();
  function ColorToBrightness(C: TColor): Integer;
  begin
    Result := C and $FF + C and $FF00 shr 8 + C and $FF0000 shr 16;
  end;
var
  i, j: Integer;
  C: TColor;
  Done: Boolean;
begin
  PaletteSize := ColorMapSize;
  if UseTransparentColor then
    Dec(PaletteSize);
  j := 0;
  for i := 0 to ColorMapSize - 1 do
  begin
    if UseTransparentColor and (i = TransparentColorIndex) then
      continue;
    Palette[j] := ColorMap[i];
    Inc(j);
  end;

  // Sort
  repeat
    Done := True;
    for j := 0 to PaletteSize - 2 do
      if ColorToBrightness(Palette[j]) < ColorToBrightness(Palette[j + 1]) then
      begin
        C := Palette[j];
        Palette[j] := Palette[j + 1];
        Palette[j + 1] := C;
        Done := False;
      end;
  until Done;
end;

function CanSetOffset(): Boolean;
begin
  Result := 0 < MaxOffset;
end;

procedure SetOffset(AOffset: Integer);
begin
  if (AOffset < 0) or (MaxOffset < AOffset) then
  begin
    DoError(msgEOffsetOutOfRange, resEOffsetOutOfRange);
    AOffset := 0;
  end;
  Offset := AOffset;
end;

type
  TOctave = 0..10; // o- ~ o9
  TNote = 0..6; // c ~ b

const
  MemberNames: array[0..15] of String = (
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'Z'
  );
  NoteNames: array[TNoteType] of array[TNote] of String = (
    ('d', 'r', 'm', 'f', 's', 'l', 'c'),
    ('c', 'd', 'e', 'f', 'g', 'a', 'b'),
    ('c', 'd', 'e', 'f', 'g', 'a', 'h')
  );
  OctaveNames: array[TOctave] of String = (
    '-', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
  );

procedure BitmapToMus();
var
  SL: TStringList;
  S: String;
  C: TColor;
  m, x, y, h: Integer;
  NoteLengthSet: Boolean;
  Notes, LastNotes: String;
  LastOctave: Integer;
  Octave: TOctave;
  Note: TNote;
  NoteCount: Integer;
  Progress, ProgressBase, ProgressRate: Integer;
  procedure FoldAddColr();
  begin
    while 75 < Length(S) do
    begin
      SL.Add(Copy(S, 1, 75));
      S := '     ' + Copy(S, 76, Length(S));
    end;
    if '     "' = S then
    begin
      SL[SL.Count - 1] := SL[SL.Count - 1] + '"';
      exit;
    end;
    SL.Add(S);
  end;
  procedure FoldAddNotes();
  begin
    while 80 < Length(S) do
    begin
      SL.Add(Copy(S, 1, 80));
      S := '    ' + Copy(S, 81, Length(S));
    end;
    SL.Add(S);
  end;
begin
  SL := TStringList.Create();
  try
    SetPalette();
    MusData.Clear();
    S := '*COLR"';
    for m := 0 to PaletteSize - 1 do
    begin
      C := Palette[m];
      if 0 < m then S := S + ' ';
      S := S + MemberNames[m] + '(' +
               Format('%.2x%.2x%.2x', [C and $FF, C and $FF00 shr 8, C and $FF0000 shr 16]) +
               ')';
    end;
    S := S + '"';
    FoldAddColr();
    SL.Add('');
    SL.Add('*MARK""');
    S := '@';
    for m := 0 to PaletteSize - 1 do S := S + MemberNames[m];
    S := S + ' T0';
    SL.Add(S);
    SL.Add('\=%%' + IntToStr(Tempo));
    SL.Add('*FING"v0?0p0q0y/' + NoteTypeNames[NoteType] + '"');
    SL.Add('');
    ProgressBase := SourceBitmap.Width * PaletteSize;
    Progress := 0;
    for m := 0 to PaletteSize - 1 do
    begin
      C := Palette[m];
      S := '#' + MemberNames[m] + '0 ';
      NoteLengthSet := False;
      LastNotes := '';
      for x := 0 to SourceBitmap.Width - 1 do
      begin
        Inc(Progress);
        ProgressRate := Progress * 100 div ProgressBase;
        if Assigned(OnB2MProgress) then OnB2MProgress(ProgressRate);
        if Assigned(B2MProgressProc) then B2MProgressProc(ProgressRate);
        LastOctave := -1;
        Notes := '';
        NoteCount := 0;
        for y := 0 to SourceBitmap.Height - 1 do
        begin
          if C <> SourceBitmap.Canvas.Pixels[x, SourceBitmap.Height - 1 - y] then continue;
          h := y + MaxOffset - Offset;
          Octave := h div 7;
          Note := h mod 7;
          if Octave <> LastOctave then
          begin
            Notes := Notes + 'o' + OctaveNames[Octave];
            LastOctave := Octave;
          end;
          Notes := Notes + NoteNames[NoteType][Note];
          Inc(NoteCount);
        end;
        if 0 = NoteCount then
          S := S + '_'
        else if 1 = NoteCount then
          if Notes = LastNotes then
            S := S + Copy(Notes, 3, 1)
          else
            S := S + Notes
        else // 1 < NoteCount
          if Notes = LastNotes then
            S := S + ','
          else
            S := S + '[' + Notes + ']';
        LastNotes := Notes;
        if not NoteLengthSet then
        begin
          S := S + NoteLength;
          NoteLengthSet := True;
        end;
      end;
      FoldAddNotes();
//      SL.Add(S);
    end;

    SL.Add('');
    SL.Add('%');
    SL.Add('*FING"v100o4"');
    SL.Add('*MARK""');
    MusData.Assign(SL);
  finally
    SL.Free();
  end;
end;

procedure SaveMus();
begin
  if FileExistsUTF8(MusPath) then
    DoWarning(Format(msgWFileExists, [MusPath]),
              Format(resWFileExists, [MusPath]));
  MusData.SaveToFile(UTF8ToSys(MusPath));
end;

procedure WriteMus();
var
  i: Integer;
begin
  for i := 0 to MusData.Count - 1 do
    Writeln(MusData[i]);
end;

initialization
  optTransparentColorSet := False;
  optHelpSet := False;
  optNoteLengthSet := False;
  optOutputFileSet := False;
  optTopOffsetSet := False;
  optTempoSet := False;
  optStdoutSet := False;
  optNoUseTransparentColorSet := False;
  optUseTransparentColorSet := False;
  optVersionSet := False;
  optVerboseSet := False;
  optNoWizardSet := False;
  optWithWizardSet := False;
  optNoteTypeSet := False;
  optBitmapPathSet := False;
  optBitmapPathsSet := False;
  optErrorSet := False;
  optTransparentColorArg := '';
  optNoteLengthArg := '';
  optOutputFileArg := '';
  optTopOffsetArg := '';
  optTempoArg := '';
  optNoteTypeArg := '';
  optBitmapPathArg := '';

  Verbose := False;
  UseWizard := False;
  BitmapPath := '';
  SourceBitmap := TBitmap.Create();
  ColorMapSize := 0;
  UseTransparentColor := False;
  TransparentColorIndex := 0;
  PaletteSize := 0;
  MaxOffset := 0;
  Offset := 0;
  NoteType := ntx1;
  NoteLength := '32';
  Tempo := 125;
  MusData := TStringList.Create();
  MusPath := '';
  OnB2MProgress := nil;
  B2MProgressProc := nil;

finalization
  SourceBitmap.Free();
  MusData.Free();
end.

