program bmp2mus;

// bmp2mus.lpr
//
//     Copyright(c) 2009-2011 PT2K <PT2K@d8.dion.ne.jp>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  Interfaces, Forms, LResources, Classes, SysUtils, FileUtil, Graphics,
  Translations, gettext, getopts,
  Wizard, ResStrs, B2MConv;

{$IFDEF WINDOWS}{$R bmp2mus.rc}{$ENDIF}

const
  VersionStr = '1.0.1';
  CopyrightYear = '2009-2011';

{
-c --transparent-color=X,Y|RGB
-h --help
-l --note-length=STR
-o --output-file=NAME
-O --top-offset=INT
-p --tempo=INT
-s --stdout
-t --no-use-transparent-color
-T --use-transparent-color
-v --version
-V --verbose
-w --no-wizard
-W --with-wizard
-x --note-type=0|1|2
}

type
  TB2MLongOptions = (
    loTransparentColor,
    loHelp,
    loNoteLength,
    loOutputFile,
    loTopOffset,
    loTempo,
    loStdout,
    loNoUseTransparentColor,
    loUseTransparentColor,
    loVersion,
    loVerbose,
    loNoWizard,
    loWithWizard,
    loNoteType,
    loEnd
  );

const
  LongOptionNames: array[TB2MLongOptions] of String = (
    'transparent-color',
    'help',
    'note-length',
    'output-file',
    'top-offset',
    'tempo',
    'stdout',
    'no-use-transparent-color',
    'use-transparent-color',
    'version',
    'verbose',
    'no-wizard',
    'with-wizard',
    'note-type',
    ''
  );
  LongOptionHasArgs: array[TB2MLongOptions] of Integer = (
    1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0
    // 0: No_Argument
    // 1: Required_Argument
    // 2: Optional_Argument
  );
  LongOptionValues: array[TB2MLongOptions] of Char = (
    'c', 'h', 'l', 'o', 'O', 'p', 's', 't', 'T', 'v', 'V', 'w', 'W', 'x', #0
  );
  ShortOptions = 'c:hl:o:O:p:stTvVwWx:';

procedure ParseOptions();
var
  o: TB2MLongOptions;
  c: char;
  OptionIndex: Longint;
  LongOptions: array[TB2MLongOptions] of TOption;
  BitmapPaths: TStringList;
begin
  for o := Low(TB2MLongOptions) to High(TB2MLongOptions) do
  begin
    LongOptions[o].Name := LongOptionNames[o];
    LongOptions[o].Has_arg := LongOptionHasArgs[o];
    LongOptions[o].Flag := nil;
    LongOptions[o].Value := LongOptionValues[o];
  end;

  OptErr := False;
  OptionIndex := 0;
  c := #0;
  repeat
    c := GetLongOpts(ShortOptions, @LongOptions[Low(TB2MLongOptions)], OptionIndex);
    case c of
      'c': begin
             optTransparentColorSet := True;
             optTransparentColorArg := OptArg;
           end;
      'h': optHelpSet := True;
      'l': begin
             optNoteLengthSet := True;
             optNoteLengthArg := OptArg;
           end;
      'o': begin
             optOutputFileSet := True;
             optOutputFileArg := OptArg;
           end;
      'O': begin
             optTopOffsetSet := True;
             optTopOffsetArg := OptArg;
           end;
      'p': begin
             optTempoSet := True;
             optTempoArg := OptArg;
           end;
      's': optStdoutSet := True;
      't': optNoUseTransparentColorSet := True;
      'T': optUseTransparentColorSet := True;
      'v': optVersionSet := True;
      'V': optVerboseSet := True;
      'w': optNoWizardSet := True;
      'W': optWithWizardSet := True;
      'x': begin
             optNoteTypeSet := True;
             optNoteTypeArg := OptArg;
           end;
      '?', ':': optErrorSet := True;
    end;
  until c = EndOfOptions;

  BitmapPaths := TStringList.Create();
  try
    while OptInd <= ParamCount do
    begin
      BitmapPaths.Add(SysToUTF8(ParamStr(OptInd)));
      Inc(OptInd);
    end;
    if 0 < BitmapPaths.Count then
    begin
      optBitmapPathSet := True;
      optBitmapPathArg := BitmapPaths[0];
    end;
    optBitmapPathsSet := 1 < BitmapPaths.Count;
  finally
    BitmapPaths.Free();
  end;
end;

function AnalyzeOptions(): Boolean;
  function SetNoteType(): Boolean;
  var
    n: TNoteType;
  begin
    Result := False;
    NoteType := ntx1;
    for n := Low(TNoteType) to High(TNoteType) do
      if NoteTypeNames[n] = 'x' + optNoteTypeArg then
      begin
        NoteType := n;
        Result := True;
        break;
      end;
  end;
  function SetTransparentColorIndex(): Boolean;
  var
    L: Longint;
    C: TColor;
    p, x, y: Integer;
    function ColorToIndex(): Boolean;
    var
      i: Integer;
    begin
      TransparentColorIndex := -1;
      for i := 0 to ColorMapSize - 1 do
        if C = ColorMap[i] then
        begin
          TransparentColorIndex := i;
          break;
        end;
      Result := 0 <= TransparentColorIndex;
    end;
  begin
    Result := False;
    // RGB
    try
      L := StrToInt('$' + optTransparentColorArg);
      if L < 0 then exit;
      C := L and $00FF0000 shr 16 + L and $0000FF00 + L and $000000FF shl 16;
      Result := ColorToIndex();
      exit;
    except
      on EConvertError do; // go next
    end;
    // Pixel
    p := Pos(',', optTransparentColorArg);
    if 0 = p then exit;
    try
      x := StrToInt(Copy(optTransparentColorArg, 1, p - 1));
      y := StrToInt(Copy(optTransparentColorArg, p + 1, Length(optTransparentColorArg)));
    except
      on EConvertError do exit;
    end;
    if (x < 0) or (y < 0) then exit;
    C := SourceBitmap.Canvas.Pixels[x, y];
    Result := ColorToIndex();
  end;
begin
  Result := False;

  Verbose := optVerboseSet;
  if not optBitmapPathSet then
    UseWizard := True
  else
    UseWizard := optWithWizardSet;
  if optErrorSet then
  begin
    DoError(msgEWrongOptions, resEWrongOptions);
    exit;
  end;
  if optVersionSet or optHelpSet then
  begin
{$IFNDEF WINDOWS}
    if not optWithWizardSet then
      UseWizard := False;
{$ENDIF}
    if optVersionSet then
      DoInformation(Format(msgIVersion, [VersionStr, CopyrightYear]),
                    Format(resIVersion, [VersionStr, CopyrightYear]));
    if optHelpSet then
      DoInformation(msgIHelp, resIHelp);
    exit;
  end;
  UseTransparentColor := optUseTransparentColorSet;
  if optBitmapPathSet then
    if not SetBitmapPath(optBitmapPathArg) then
      if not UseWizard then exit;
  if optTransparentColorSet then
    if not SetTransparentColorIndex() then
      if not optBitmapPathSet then
      begin
        DoError(msgEOptCNeedsBitmap, resEOptCNeedsBitmap);
        exit;
      end
      else
      begin
        DoError(Format(msgEInvalidColor, [optTransparentColorArg]),
                Format(resEInvalidColor, [optTransparentColorArg]));
        exit;
      end;
  if optNoteTypeSet then
    if not SetNoteType() then
    begin
      DoError(Format(msgEInvalidNoteType, [optNoteTypeArg]),
              Format(resEInvalidNoteType, [optNoteTypeArg]));
      exit;
    end;
  if optNoteLengthSet then
    NoteLength := optNoteLengthArg;
  if optTempoSet then
  try
    Tempo := StrToInt(optTempoArg);
    if (Tempo < 1) or (999 < Tempo) then
      raise EConvertError.Create('');
  except
    on EConvertError do
    begin
      DoError(Format(msgEInvalidTempo, [optTempoArg]),
              Format(resEInvalidTempo, [optTempoArg]));
      exit;
    end;
  end;
  if optTopOffsetSet then
  try
    SetOffset(StrToInt(optTopOffsetArg));
  except
    on EConvertError do
    begin
      DoError(Format(msgEInvalidOffset, [optTopOffsetArg]),
              Format(resEInvalidOffset, [optTopOffsetArg]));
      exit;
    end;
  end;
{$IFDEF WINDOWS}
  if optStdoutSet then
  begin
    DoWarning(msgWStdoutIgnored, resWStdoutIgnored);
    optStdoutSet := False;
  end;
{$ENDIF}

  // Check conflicts
  if optNoWizardSet and optWithWizardSet then
    DoWarning(Format(msgWOptionConflicts, ['-w', '-W']),
              Format(resWOptionConflicts, ['-w', '-W']));
  if optBitmapPathsSet then
    DoWarning(msgWTooMuchFiles, resWTooMuchFiles);
  if optNoUseTransparentColorSet and optUseTransparentColorSet then
    DoWarning(Format(msgWOptionConflicts, ['-t', '-T']),
              Format(resWOptionConflicts, ['-t', '-T']));
  if UseWizard and optStdoutSet then
    DoWarning(msgWStdoutIncompatible, resWStdoutIncompatible);
  if ForceUseTransparentColor() and optNoUseTransparentColorSet then
    DoWarning(msgWForceUseTransparentColor, resWForceUseTransparentColor);
  if optNoUseTransparentColorSet and optTransparentColorSet then
    DoWarning(Format(msgWOptionConflicts, ['-c', '-t']),
              Format(resWOptionConflicts, ['-c', '-t']));

  Result := True;
end;

procedure GetLanguageIDsFromEnvironmentVariable(var Lang, FallbackLang: string);
begin
  lang := GetEnvironmentVariable('LC_ALL');
  if Length(lang) = 0 then
  begin
    lang := GetEnvironmentVariable('LC_MESSAGES');
    if Length(lang) = 0 then
    begin
      lang := GetEnvironmentVariable('LANG');
      if Length(lang) = 0 then
        exit;   // no language defined via environment variables
    end;
  end;
  FallbackLang := Copy(lang, 1, 2);
end;

procedure TranslateLCL();
var
  PoPath, Lang, FallbackLang: String;
begin
  PoPath := AppendPathDelim(ProgramDirectory() + 'po') + 'bmp2mus.%s.po';
  Lang := '';
  FallbackLang := '';
  GetLanguageIDsFromEnvironmentVariable(Lang, FallbackLang);
  Translations.TranslateUnitResourceStrings('ResStrs', PoPath, Lang, FallbackLang);
end;

begin
  {$I bmp2mus.lrs}
  Application.Initialize();
  TranslateLCL();
  ParseOptions();
  if not AnalyzeOptions() then exit;
  if UseWizard then
  begin
    Application.CreateForm(TB2MWizard, B2MWizard);
    Application.Run();
  end
  else
  begin
    BitmapToMus();
    if optStdOutSet then
      WriteMus()
    else
      SaveMus();
  end;
end.

